import string

def name(s):
    if type(s) is not str:
        raise TypeError(f"A name must be a string; {s.__repr__()} is not! ")
    invalid_chars = set(s) - set(string.digits + string.ascii_lowercase + "_")
    if (len(invalid_chars) or (s[0] in string.digits)):
        raise ValueError(f"The value {s.__repr__()} is not a valid name! ")
    return s
def datatype(t):
    return {
        int: 'INTEGER',
        str: 'TEXT',
        float: 'REAL',
    }[t]




from . import fmt

def where(equals_questionmark, *, is_null=[], is_not_null=[]):
    conditions = list()
    conditions += [f"{fmt.name(x)} = ?" for x in equals_questionmark]
    conditions += [f"{fmt.name(x)} IS NULL" for x in is_null]
    conditions += [f"{fmt.name(x)} IS NOT NULL" for x in is_not_null]
    if not len(conditions):
        return "WHERE 1"
    cmd = "WHERE (("
    cmd += ") AND (".join(conditions)
    cmd += "))"
    return cmd


import sys
import pandas as pd
from . import fmt
from . import cmd
from .Where import Where




def _cursor_decorator(oldfunc):
    def newfunc(*, cursor, **kwargs):
        command, values = oldfunc(**kwargs)
        _main(
            cursor=cursor,
            command=command,
            values=values,
        )
    return newfunc



def _where_decorator(oldfunc):
    def newfunc(*, where, **kwargs):
        where = Where(row=where)
        return oldfunc(where=where, **kwargs)
    return newfunc




def _main(*, cursor, command, values):
    command = str(command)
    values = list(values)
    try:
        cursor.execute(command, values)
    except:
        raise Exception(f"cursor.execute({command.__repr__()}, {values}) failed! ")    





def insert(*, cursor, table, row):
    row = dict(row)
    command = "INSERT INTO "
    command += fmt.name(table)
    command += "("
    command += ", ".join(fmt.name(k) for k in row.keys())
    command += ")"
    command += " VALUES "
    command += "("
    command += ", ".join('?' for v in row.values())
    command += ");"
    _main(
        cursor=cursor,
        command=command,
        values=row.values(),
    )




@_where_decorator
def update(*, cursor, table, row, where):
    #where = dict(where)
    row = dict(row)
    command = "UPDATE "
    command += fmt.name(table)
    command += " SET "
    command += ", ".join(f"{fmt.name(k)} = ?" for k in row.keys())
    command += " "
    command += where.command
    command += ";"
    #print('table', table, file=sys.stderr)
    #print('row', row, file=sys.stderr)
    #print('where', where, file=sys.stderr)
    #print(command, list(row.values()) + where.values, file=sys.stderr)
    #print(file=sys.stderr)
    _main(
        cursor=cursor,
        command=command,
        values=list(row.values()) + where.values,
    )




@_where_decorator
def select(*, cursor, table, columns, where={}):
    columns = list(columns)
    if len(columns) == 0:
        raise ValueError("The columns-list is not allowed to be empty! ")
    command = "SELECT "
    if '*' in columns:
        if len(columns) != 1:
            raise ValueError("The columns-list is invalid! ")
        command += '*'
    else:
        command += ", ".join(fmt.name(c) for c in columns)
    command += " FROM "
    command += fmt.name(table)
    command += " "
    command += where.command#cmd.where(where.keys())
    command += ";"
    _main(
        cursor=cursor,
        command=command,
        values=where.values,
    )
    if '*' in columns:
        columns = [descriptor[0] for descriptor in cursor.description]
    return pd.DataFrame(data=cursor.fetchall(), columns=columns)







def all_tables(*, cursor):
    df_master = select(
        cursor=cursor,
        table='sqlite_master',
        columns=['name'],
        where={'type':'table'},
    )
    ans = dict()
    for table in df_master['name']:
        ans[table] = select(
            cursor=cursor,
            table=table,
            columns=['*'],
            where={},
        )
    return ans











import pandas as pd
from . import fmt

class Where:
    def __str__(self):
        return self.command
    @property
    def command(self):
        return self._command
    @property
    def values(self):
        return list(self._values)
    def __init__(self, /, row, *, is_not_null=[]):
        self._command = ""
        self._values = list()
        row = dict(row)
        is_not_null = list(is_not_null)
        conditions = list()
        for k, v in row.items():
            condition = fmt.name(k)
            if type(v) in (tuple, list, set, frozenset):
                options = list(set(v))
                condition += " IN ("
                condition += ', '.join(['?'] * len(options))
                condition += ")"
                self._values += options
            elif pd.isna(v):
                condition += " IS NULL"
            else:
                condition += " = ?"
                self._values.append(v)
            conditions.append(condition)
        for k in is_not_null:
            condition = fmt.name(k)
            condition += " IS NOT NULL"
            conditions.append(condition)
        if len(conditions) == 0:
            self._command = "WHERE 1"
        elif len(conditions) == 1:
            self._command = f"WHERE ({conditions[0]})"
        else:
            self._command = "WHERE (("
            self._command += ") AND (".join(conditions)
            self._command += "))"
            
